// REEF BLOWER

local PULLRANGE = 256*FRACUNIT
local PULLHEIGHT = 64*FRACUNIT

//
// P_AddBlowerItem
//
// Adds an item to the reef blower
//
local function P_AddBlowerItem(sponge, item)
	
	local mobj = {}
	local things = sponge.ssvars.ammo and sponge.ssvars.ammo["things"] and sponge.ssvars.ammo["things"][SS_POWERUP_BLOWER] or {}
	local oldtype = item.extravalue3
	local maxammo = supersponge_powerup_info[SS_POWERUP_BLOWER].maxammo

	// Already full?!
	if (things != {} and #things >= maxammo) then CONS_Printf(sponge, "Your reef blower is full!") return end	
	
	// Reset item type back to default if not already.
	if (item.type == MT_SUPERSPONGE_FAKEMISSILE and oldtype) then
		item.type = oldtype
	end
	
	mobj = {
		type = item.type,
		sprite = item.sprite,
		flags = item.flags,
		color = item.color,
		colorized = item.colorized,
		scale = item.scale,
		deathstate = item.info.deathstate
	}
	
	table.insert(sponge.ssvars.ammo.things[SS_POWERUP_BLOWER], mobj)	
end

//
// P_RemoveBlowerItem
//
// Removes the last item from the reef blower
//
local function P_RemoveBlowerItem(sponge)

	local things = sponge.ssvars.ammo and sponge.ssvars.ammo["things"] and sponge.ssvars.ammo["things"][SS_POWERUP_BLOWER] or {}
	
	// Nothing inside.
	if (things == {}) then CONS_Printf(sponge, "Your reef blower is empty!") return end
	
	for k,v in ipairs(things) do

		if (k != #things) then
			continue // not the last item in set
		end

		if (v == nil) then
			continue // non-existent
		end
		
		table.remove(things, k)
		
		return true
	end
	
	return false
end

//
// P_GetLastBlowerItem
//
// Returns the last item from the reef blower.
//
local function P_GetLastBlowerItem(sponge)

	local things = sponge.ssvars.ammo["things"] and sponge.ssvars.ammo["things"][SS_POWERUP_BLOWER] or {}
	
	// Nothing inside.
	if (things == {}) then CONS_Printf(sponge, "Your reef blower is empty!") return nil end
	
	for k,v in ipairs(things) do
		
		if (k != #things) then
			continue // not the last item in set
		end
		
		if (v == nil) then
			continue // non-existent
		end
		
		return v // Found an item.
	end
	
	CONS_Printf("Could not find any valid items...")
	
	return nil
end

//
// P_GetBlowerCount
//
// Returns the count of items from a Reef Blower.
//
rawset(_G, "P_GetBlowerCount", function(sponge)
	local count = 0
	local things = sponge.ssvars.ammo and sponge.ssvars.ammo["things"] and sponge.ssvars.ammo["things"][SS_POWERUP_BLOWER] or {}
	
	if (things == {} or things == nil) then
		return 0 // empty
	end
	
	for k,v in ipairs(things) do
	
		if (v == nil) then
			continue
		end
		
		count = $ + 1
	end
	
	return count
end)

local function P_SetVacuumStatus(sponge, status)
	sponge.ssvars.blower.status = status
end	

local function P_GetVacuumStatus(sponge)
	return sponge.ssvars.blower.status
end	

//
// P_RemoveBlower
//
// Removes the blower and "ejects" any obstructing objects in front of it.
//
rawset(_G, "P_RemoveBlower", function(sponge)

	local blower = sponge.ssvars.blower
	local obstructor = sponge.ssvars.blower.obstructor or nil
	local cooldown = 10
	
	if (obstructor and obstructor.valid) then
	
		// Remove MF_NOCLIPHEIGHT for players.
		if (obstructor.player) then
			obstructor.flags = $ & ~MF_NOCLIPHEIGHT
		end
	
		obstructor.blocked_by_wall = false // Set wall flag
		obstructor.rbplayer = sponge.mo // Get the last Spongebob who previously launched you.
		obstructor.rbangle = sponge.drawangle // Get the last angle you were launched at.
		obstructor.rblaunchtime = 8 // Set pain time
		obstructor.rbflags = RBF_LAUNCHED // "Launch" whatever is in the way.	
		
		P_MoveOrigin(obstructor, sponge.mo.x + P_ReturnThrustX(sponge.mo, sponge.drawangle, -obstructor.oldradius), 
							     sponge.mo.y + P_ReturnThrustY(sponge.mo, sponge.drawangle, -obstructor.oldradius), 
								 sponge.mo.z)
		
		P_InstaThrust(obstructor, sponge.drawangle, FRACUNIT)
		
		obstructor.height = obstructor.oldheight // Restore previous height
		obstructor.radius = obstructor.oldradius // Restore previous radius
	
		cooldown = TICRATE // After launching your obstructor, the blower will be unsuable for one second.
	end

	sponge.ssvars.blower.obstructor = nil
	sponge.ssvars.blower.status = RB_STATUS_OFF
	sponge.ssvars.blower.cooldown = cooldown
	
	// Remove the blower hitbox if it still exists.
	if (blower.hitbox and blower.hitbox.valid) then
		P_RemoveMobj(blower.hitbox)
	end
	
	// If the reef blower was making noise, emit the blower stop sound.
	if (sponge.ssvars.blower.noise) then	
		S_StartSound(sponge.mo, sfx_rbstop)
	end
	
	sponge.ssvars.blower.noise = 0
end)

//
// P_ExplodeBlower
//
// Same thing but with explosions!
//
local function P_ExplodeBlower(sponge)

	local blower = sponge.ssvars.blower
	local obstructor = sponge.ssvars.blower.obstructor or nil
	local dust = P_SpawnMobjFromMobj(sponge.mo, 0, 0, 0, MT_TNTDUST)
	local bomb = P_SpawnMobjFromMobj(sponge.mo, 0, 0, 0, MT_RAY)
	
	// The blower is now gone, so remove it immediately.
	P_RemoveSSPowerup(sponge, SS_POWERUP_BLOWER, false)
	
	sponge.ssvars.blower.obstructor = nil
	sponge.ssvars.blower.status = RB_STATUS_OFF
	sponge.ssvars.blower.cooldown = cooldown
	
	// Removethe blower hitbox if it still exists.
	if (blower.hitbox and blower.hitbox.valid) then
		P_RemoveMobj(blower.hitbox)
	end
	
	dust.destscale = $*4
	
	S_StartSound(bomb.subsector.sector, sfx_bebomb) // Cue the explosion noise.
	
	// First, kill the obstructor.
	if (obstructor and obstructor.valid) then
		P_DamageMobj(obstructor, nil, nil, 0, DMG_INSTAKILL)
	end	
	
	// Now, kill the player.
	if (sponge.mo and sponge.mo.valid) then
		P_DamageMobj(sponge.mo, nil, nil, 0, DMG_INSTAKILL)
	end
	
	P_RadiusAttack(bomb, sponge.mo, 512*FRACUNIT, 100, true)
end

//
// P_LaunchVacuumItem
//
// Fires a vacuum item or obstructing object from the player's blower.
//
rawset(_G, "P_LaunchVacuumItem", function(sponge)

	local item
	local item_type
	local rocket
	local obstructor

	// Something is blocking the reef blower
	if (P_GetVacuumStatus(sponge) == RB_STATUS_STUCK) then
	
		local angle = sponge.ssvars.camchase and sponge.drawangle or sponge.mo.angle
		local aiming = sponge.cmd.aiming << 16
	
		sponge.weapondelay = 1
	
		obstructor = sponge.ssvars.blower.obstructor

		P_RemoveBlower(sponge) // Remove the blower before launching the obstructor
		
		if not (obstructor and obstructor.valid and obstructor.health > 0) then
			return // already dead or non-existent
		end
	
		// Remove MF_NOCLIPHEIGHT for players.
		if (obstructor.player) then
			obstructor.flags = $ & ~MF_NOCLIPHEIGHT
		end
	
		obstructor.blocked_by_wall = false // Set wall flag
		obstructor.rbplayer = sponge.mo // Get the last Spongebob who previously launched you.
		obstructor.rbangle = angle // Get the last angle you were launched at.
		obstructor.rblaunchtime = 8 // Set pain time
		obstructor.rbflags = RBF_LAUNCHED // "Launch" whatever is in the way.
		
		P_MoveOrigin(obstructor, sponge.mo.x + P_ReturnThrustX(sponge.mo, sponge.drawangle, -obstructor.oldradius), 
							     sponge.mo.y + P_ReturnThrustY(sponge.mo, sponge.drawangle, -obstructor.oldradius), 
								 sponge.mo.z)
		
		obstructor.height = obstructor.oldheight
		obstructor.radius = obstructor.oldradius
								 
		P_InstaThrust(obstructor, sponge.drawangle, 40*FRACUNIT) // Thrust 40*FRACUNITS forwards.
		P_SetObjectMomZ(obstructor, P_ReturnThrustY(obstructor, aiming, 24*FRACUNIT), true) // Thrust 24*FRACUNITS vertically.

		sponge.ssvars.blower.cooldown = SS_SHOOT_COOLDOWN
		
	// Fire any stored objects if the reef blower is not empty
	elseif (P_GetBlowerCount(sponge) > 0) then
	
		local angle = sponge.ssvars.camchase and sponge.drawangle or sponge.mo.angle
	
		sponge.weapondelay = 1
		
		item = P_GetLastBlowerItem(sponge)
		
		rocket = P_SpawnSpongeMissile(sponge, item, false)

		// Boost X / Y momentum
		rocket.momx = FixedMul(40*FRACUNIT, cos(angle))
		rocket.momy = FixedMul(40*FRACUNIT, sin(angle))
		
		// Adjust to aiming view
		rocket.momx = FixedMul($, cos(sponge.aiming))
		rocket.momy = FixedMul($, cos(sponge.aiming))
		
		// Boost Z momentum
		rocket.momz = FixedMul(39*FRACUNIT, sin(sponge.aiming))
		
		// Try multiplying momentum with scale
		rocket.momx = FixedMul($, rocket.scale)
		rocket.momy = FixedMul($, rocket.scale)
		rocket.momz = FixedMul($, rocket.scale)
		
		P_RemoveBlowerItem(sponge)
		
		sponge.ssvars.blower.cooldown = SS_SHOOT_COOLDOWN/2
	end
end)

// P_UseReefBlower
//
// Handles the usage of the reef blower.
//
rawset(_G, "P_UseReefBlower", function(sponge)
	
	if (sponge.mo.state == S_PLAY_JUMP)
	or (sponge.mo.state == S_PLAY_ROLL) then
		return // jumping or rolling
	end
	
	if (sponge.ssvars.meleetimer > 0 or sponge.ssvars.blower.cooldown > 0) then
		return // melee timer or cooldown timer is active
	end
	
	// Spawn the blower hitbox, if it doesn't already exist
	if not (sponge.ssvars.blower.hitbox and sponge.ssvars.blower.hitbox.valid) then
		sponge.ssvars.blower.hitbox = P_SpawnMobjFromMobj(sponge.mo, 0, 0, 0, MT_SUPERSPONGE_VACBOX)	
		sponge.ssvars.blower.hitbox.target = sponge.mo
	end
	
	sponge.ssvars.blower.status = RB_STATUS_ON // set status to ON
end)

//
// P_IsMobjVacuumable
//
// Checks for whether mobj can actually be vacuumed at all.
//
rawset(_G, "P_IsMobjVacuumable", function(mobj)

	local suck = false
	local whitelist = supersponge_blower_whitelist
	local is_valid_target = false
	local is_vacuumable = false
	local is_collectible = false
	
	is_valid_target = (mobj.flags & (MF_ENEMY|MF_MONITOR|MF_MISSILE)) or false
	is_vacuumable = whitelist[mobj.type] != nil and whitelist[mobj.type] >= 1 or false
	is_collectible = supersponge_collectible_list[mobj.type] != nil and supersponge_collectible_list[mobj.type] or false
	
	// These objects can never be vacuumed.
	if (whitelist[mobj.type] != nil and whitelist[mobj.type] <= 0) then
		is_valid_target = false
	end
	
	// Never target shockwaves.
	if (mobj.type == MT_SHOCKWAVE) then
		is_valid_target = false
	end
	
	// Just go with the assumption that golden monitors are extremely heavy.
	if (mobj.flags & MF_MONITOR and mobj.is_golden) then
		is_valid_target = false
	end
	
	// Only suck enemies, monitors, and missiles or whitelisted objects.
	if (mobj and mobj.valid and mobj.health > 0 and (is_valid_target or is_vacuumable or is_collectible)) then
		suck = true
	end
	
	return suck
end)

local gBlowerList = {} // global list of affected things

// 
// UpdateBlowerList
//
// Adds mobj to the global blower iterator
//
local function UpdateBlowerList(mobj)

	local in_list = false
	
	if not (#gBlowerList) then return false end // table is empty 
	
	// iterate through the list and check if mobj is already in it.
	for k,v in ipairs(gBlowerList) do
	
		// the mobj is already in the list
		if (v != mobj) then
			continue
		end
		
		in_list = true
		
		break
	end
		
	// already got mobj
	if (in_list) then 
		return true
	end
	
	table.insert(gBlowerList, mobj)
	
	return true
end

//
// P_SSVacuumCheck
//
// Checks to see if the target will either obstruct or be ignored by the blower.
//
local function P_SSVacuumCheck(player, target)

	local whitelist = supersponge_blower_whitelist

	// The target does not obstruct the blower, so it's okay to vacuum.
	if (whitelist[target.type] == nil or whitelist[target.type] == 1) then
		return true
	end
	
	// These objects can never be vacuumed.
	if (whitelist[target.type] <= 0) then
		return false
	end

	player.ssvars.blower.obstructor = target // set obstructor
	player.ssvars.blower.obstructor.oldheight = target.height // set old height
	player.ssvars.blower.obstructor.oldradius = target.radius // set old radius
	player.ssvars.blower.obstructor.rbflags = RBF_OBSTRUCT // set flag
	
	// For enemies, set their state to spawnstate and set reactiontime.
	if (target.flags & MF_ENEMY) then
		target.state = target.info.spawnstate
		target.reactiontime = TICRATE
	end

	// Set obstructor's height and radius
	if (player.ssvars.blower.hitbox and player.ssvars.blower.hitbox.valid) then
		target.height = player.ssvars.blower.hitbox.height
		target.radius = player.ssvars.blower.hitbox.radius
	end
	
	P_SetVacuumStatus(player, RB_STATUS_STUCK) // set status
	
	return false
end

//
// P_DoBlowerPop
//
// Handles monitor destruction by the player's blower.
//
local function P_DoBlowerPop(blower, monitor)

	P_KillMobj(monitor, blower, blower.target)
	
	monitor.state = S_BOX_POP2 // Set state to S_BOX_POP2
	monitor.sprite = SPR_NULL // Hide monitor sprite.
	
	// If the monitor can respawn outside of Singleplayer / Coop, move it back to it's origin for the next monitor.
	if not G_CoopGametype() then
		P_SetOrigin(monitor, monitor.origin.x, monitor.origin.y, monitor.origin.z)
	end
	
	monitor.momx = 0
	monitor.momy = 0
	monitor.momz = 0
end

//
// P_SSVacuumCapture
//
// Adds the specified target to blower, if the blower is not full.
//
rawset(_G, "P_SSVacuumCapture", function(blower, target)

	local type
	local sponge
	
	type = target.type
	sponge = blower.target.player
	
	// No valid player...
	if not (sponge and sponge.valid) then
		return false
	end
	
	// Check if the reef blower is full before adding!
	if (P_GetBlowerCount(sponge) >= supersponge_powerup_info[SS_POWERUP_BLOWER].maxammo) then
		return false
	end
	
	// If our target was the Egg Shield, break it.
	if (target.type == MT_EGGSHIELD) then
	
		P_KillMobj(target, blower.target, blower.target, 0) 
		
		S_StartSound(target, sfx_wbreak)
		
		return false 
	end
	
	// Spawn collectible from blower's source.
	if supersponge_collectible_list[target.type] then P_SpawnMobjFromMobj(blower.target, 0, 0, 0, type) end	
	
	// Kill monitors.
	if (target.flags & MF_MONITOR) then P_DoBlowerPop(blower, target) return true end
	
	// Special handling for big targets!
	if not P_SSVacuumCheck(sponge, target) then
		return false
	end
	
	// Add enemy / missile to the reef blower list
	if (target.flags & (MF_ENEMY|MF_MISSILE) or target.type == MT_ENERGYBALL) then
		P_AddBlowerItem(blower.target.player, target)
	end
	
	if (target.type == MT_PYREFLY) then
		P_RemovePyreFly(target) // Specialized removal for Pyreflies
	elseif (target.type == MT_GSNAPPER or target.type == MT_CRUSHSTACEAN or target.type == MT_BANPYURA) then
		P_RemoveSegmentedEnemy(target) // Specialized removal for multi-segmented enemies.
	else
		P_RemoveMobj(target)
	end
end)

// 
// P_SSVacuumPull
//
// Otherwise known as...not P_Telekinesis!
//
local function P_SSVacuumPull(blower, speed, target)
	
	local an = R_PointToAngle2(blower.x, blower.y, target.x, target.y)
	local dist = P_AproxDistance(P_AproxDistance(blower.x - target.x, blower.y - target.y), blower.z - target.z)
	local ignore = false
	
	if (dist < 1) then
		dist = 1
	end
	
	speed = FixedMul(speed, speed/FixedInt(dist)) + 1
	
	// The player is trying to escape the blower
	if (target.player and dist > MELEERANGE and P_GetPlayerControlDirection(target.player) == 2) then
		// Allow the player to get away, if they are jumping or spinning.
		if (target.player.pflags & PF_JUMPED) or (target.player.pflags & (PF_SPINDOWN|PF_SPINNING)) then
			speed = $ * -1
		end
	end
	
	P_Thrust(target, an, -speed)
	
	// If your target is way above you or below you, try pulling them towards your height.
	if (blower.z > target.z + target.height) then
		P_SetObjectMomZ(target, speed, true)
	elseif (blower.z < target.z - target.height) then
		P_SetObjectMomZ(target, -speed, true)
	end

	if (target.type == MT_GOLDBUZZ or target.type == MT_REDBUZZ) then
		target.tics = $ + 8
	end
end

//
// SpongeVacuumSearch
//
// Looks for valid targets and slowly pulls them towards the blower.
//
local function SpongeVacuumSearch(blower, target)

	local sponge
	local span
	local height
	local dist
	local angle
	local pullrange
	local pullheight
	local collectible

	span = ANGLE_90
	sponge = blower.target.player
	
	if not (target and target.valid) then return end // Target does not exist
	
	if (target == blower or target == blower.target) return end // Don't target your owner or yourself!
	
	if not P_IsMobjVacuumable(target) then return end // Can't be vacuumed...
	
	pullrange = PULLRANGE
	pullheight = PULLHEIGHT
	
	// Check for other players who may be present in the vicinity.
	if (target and target.player and target.player.valid) then
	
		// check if ss_vacuumplayers allows players to be vacuumed at all
		if not CV_FindVar("ss_vacuumplayers").value then
			return
		end
		
		// The player is currently being carried by something other than a player.
		if (target.player.powers[pw_carry] > 0) then
			return
		end
		if (P_IsSuperSponge(target.skin) and target.player.ssvars.powerups[SS_POWERUP_BLOWER]) then
			return // Do NOT pull in other players who are currently using a Reef Blower.
		end
		if (target.player.powers[pw_super]) then
			pullrange = $/2 // The pull range for super players is reduced by half.
		end
	end

	height = abs(target.z - blower.target.z)
	dist = P_AproxDistance(target.x - blower.target.x, target.y - blower.target.y)
	angle = R_PointToAngle2(blower.x, blower.y, target.x, target.y) - blower.target.player.drawangle
	collectible = supersponge_collectible_list[target.type] != nil and supersponge_collectible_list[target.type] == 1 or nil
	
	if not (dist <= pullrange) // Not in pulling distance.
	or not (height <= FixedMul(pullheight, blower.scale)) // The height distance is greater than the pull height.
	or not (angle < ANGLE_90 and angle > ANGLE_270) // The target is behind us.
	or not (P_CheckSight(blower, target)) then // can't see the target
		
		if (target.sucker == blower) then
			target.sucker = nil 
		end	
		
		if (collectible) then
			target.momx = 0
			target.momy = 0
			target.momz = 0
		end

		return
	end
	
	// Reel' em in!
	if (target.sucker != blower)
		target.sucker = blower
	end
	
	// For collectibles, remove the MF_NOGRAVITY and MF_NOCLIPHEIGHT flag.
	if (collectible) then
		target.flags = $ & ~(MF_NOGRAVITY|MF_NOCLIPHEIGHT)
	end
	
	// If the list is empty, add the target to the blower.
	if not UpdateBlowerList(target) then
		table.insert(gBlowerList, target)
	end
	
	local homedist = P_AproxDistance(P_AproxDistance(target.x - blower.x, target.y - blower.y), target.z - blower.z) // Homing distance
	local speed = target.type == MT_ENERGYBALL and 40*FRACUNIT or 8*FRACUNIT // Speed
	
	 // Continuously pull the target towards the blower if not close.
	if (homedist > blower.height * 2) then 
	
		P_SSVacuumPull(blower, speed, target)
		
		return
	end
	
	P_SSVacuumCapture(blower, target)
end

//
// P_SpongeVacuumThink
//
// Handles the search function for the Reef Blower. 
// Some code from P_LookForEnemies is used as reference.
//
rawset(_G, "P_SpongeVacuumThink", function(blower)
	
	local x1, x2, y1, y2
	
	// The reef blower's broken!
	if not (blower and blower.valid) then
		return
	end

	// Who's using the reef blower?
	if not (blower.target and blower.target.valid) then
		return
	end
	
	x1 = blower.x - PULLRANGE/2
	x2 = blower.x + PULLRANGE/2
	y1 = blower.y - PULLRANGE/2
	y2 = blower.y + PULLRANGE/2

	local blockmap = searchBlockmap("objects", SpongeVacuumSearch, blower, x1, x2, y1, y2) // Find other nearby objects and pull them in.
end)

// 
// MobjBlowerMovement
//
// A helper function to move obstructor to nozzle position
//
local function MobjBlowerMovement(mobj)
		
	local x, y, z
	local blower
	local player
	local angle
	
	if not (mobj.rbflags == RBF_OBSTRUCT and mobj.sucker and mobj.sucker.valid) then
		return // blower is removed or we're not obstructing anything
	end
	
	blower = mobj.sucker
	player = mobj.sucker.target.player
	angle = player and player.drawangle or mobj.angle
	
	// If the mobj's momz was altered, slowly set it back to 0.
	if P_MobjFlip(mobj) == -1 then
		if (mobj.momz < 0) then
			mobj.momz = $ + FRACUNIT
		else
			mobj.momz = 0
		end
	else
		if (mobj.momz > 0) then
			mobj.momz = $ - FRACUNIT
		else
			mobj.momz = 0
		end
	end
	
	x = mobj.player and blower.target.x or blower.x
	y = mobj.player and blower.target.y or blower.y
	z = mobj.player and blower.target.z or blower.z
	
	P_MoveOrigin(mobj, x, y, z + FRACUNIT)
	
	if not (mobj.player) 
		return // players handle movement very differently from regular objects
	end
		
	P_InstaThrust(mobj, angle, 32*FRACUNIT)
	P_Thrust(mobj, angle-ANGLE_90, 16*FRACUNIT)
end
addHook("MobjThinker", MobjBlowerMovement, MT_PLAYER)

//
// MobjBlowerCheck
//
// Mobj Handler for Blower
//
local function MobjBlowerCheck(mobj)
	
	local list
	local sponge
	local height
	local dist
	local blower
	local monitor
	
	if not (gamestate == GS_LEVEL) then
		return // not in a level
	end
	
	if not (#gBlowerList) then
		return // list is empty
	end
	
	list = gBlowerList
	
	// print("Number of active thinkers: " + #gBlowerList)
	
	for k,v in ipairs(gBlowerList) do
	
		local mobj = v
		
		// The mobj is either dead or non-existent.
		if not (mobj and mobj.valid and mobj.health > 0) then
			
			table.remove(gBlowerList, k)
		
			continue  // stop checking, the mobj is not being pulled anymore.	
		end
	
		// The mobj is not currently being launched, pulled, or sucked.
		if not (mobj.rblaunchtime > 0 or mobj.rbflags > 0 or mobj.sucker and mobj.sucker.valid) then
					
			// Halt all momentum for stationary collectibles.
			if (supersponge_collectible_list[mobj.type] == 1)
				mobj.momx = 0
				mobj.momy = 0
				mobj.momz = 0
			end	
			
			// Remove MF_NOCLIPTHING for BattleMod's Sparring Dummy.
			if (mobj.dummy and mobj.rblaunchtime <= 0) then
				mobj.flags = $ & ~MF_NOCLIPTHING
			end
			
			table.remove(gBlowerList, k)
			
			continue  // stop checking, the mobj is not being pulled anymore.	
		end
		
		// This timer acts a short interval before objects are allowed to hurt players while being launched.
		if (mobj.rblaunchtime) then	
			mobj.rblaunchtime = $ - 1
		end
		
		// Obstructing a Reef Blower
		if (mobj.rbflags == RBF_OBSTRUCT and mobj.sucker and mobj.sucker.valid and mobj.sucker.target and mobj.sucker.target.valid) then
			
			local blower = mobj.sucker
			local player = mobj.sucker.target.player
			local angle = player and player.drawangle
			
			// Move the object's position to the blower's nozzle
			if not (mobj.player) then
				MobjBlowerMovement(mobj)
			end
			
			// We don't want enemies to hurt the player while obstructing the blower.
			if (mobj.flags & MF_ENEMY) then
				
				// Remove MF2_FRET to prevent constant flashing.
				if (mobj.flags2 & MF2_FRET) then
					mobj.flags2 = $ & ~MF2_FRET
				end
			
				mobj.reactiontime = TICRATE
			end

			// Set variables for BattleMod's Sparring Dummy to prevent it from colliding.
			if (BattleModIsActive() and mobj.type == MT_SPARRINGDUMMY) then
				mobj.dummy = true
				mobj.pushed = false
				mobj.pushed_last = nil
				mobj.pushtics = 0
			end
			
			// Handle player animations while obstructing the blower.
			if (mobj.player) then
			
				mobj.flags = $|MF_NOCLIPHEIGHT
					
				// For players who can roll or spin, use the rolling animation
				if not (mobj.player.charflags & SF_NOJUMPSPIN) then
					mobj.state = S_PLAY_ROLL
				else
					mobj.state = S_PLAY_FALL
				end
				
				mobj.player.powers[pw_nocontrol] = 1
			end
			
			continue
		end
		
		// Being launched in mid air.
		if (mobj.rbflags == RBF_LAUNCHED) then
			
			local angle
			local spongebob
			local blocked
			local stop

			angle = mobj.rbangle != nil and mobj.rbangle or 0
			spongebob = mobj.rbplayer and mobj.rbplayer.valid or nil
		
			// Hit the ground.
			if P_IsObjectOnGround(mobj) then
				stop = true
			end
			
			// Hit a solid wall or thing.
			if (mobj.blocked_by_wall and (mobj.flags & MF_SHOOTABLE) and mobj.type != MT_PLAYER) then	
			
				P_DamageMobj(mobj, spongebob, spongebob, 1)
			
				mobj.blocked_by_wall = false
			
				mobj.rbflags = $ & ~RBF_LAUNCHED
				
				continue
			end
			
			// Stop movement for players.
			if (mobj.player) then
			
				// Stop, if the player is in pain or is currently being carried.
				if (P_PlayerInPain(mobj.player) or mobj.player.powers[pw_carry]) then
					stop = true
				end
				
				mobj.player.powers[pw_nocontrol] = 1
			end
			
			// If we landed, got hit, or hit a wall, remove the flag.
			if (stop) then		
			
				mobj.blocked_by_wall = false
			
				// Set state to S_PLAY_FALL, if not being carried.
				if (mobj.player and mobj.player.powers[pw_carry] < 0) then
					mobj.state = S_PLAY_FALL
				end
			
				mobj.rbflags = $ & ~RBF_LAUNCHED
			end
			
			continue
		end
		
		// The blower is still active.
		if not (mobj.sucker and not mobj.sucker.valid) then
			continue
		end 
			
		mobj.sucker = nil	
	end
end
addHook("ThinkFrame", MobjBlowerCheck)

//
// BlowerThinker
//
// The main thinking function for the blower.
//
local function BlowerThinker(blower)

	local obstructor

	// Who's running the blower?
	if not (blower and blower.valid and blower.target and blower.target.valid and blower.target.player) then
		return
	end
	
	local sponge = blower.target.player
	local angle = sponge.drawangle
	local ammo = P_GetBlowerCount(sponge)
	local maxammo = supersponge_powerup_info[SS_POWERUP_BLOWER].maxammo
	
	// If the blower is full or the player is currently super, stop it from running!
	if (ammo == maxammo or sponge.powers[pw_super] > 0)
		return P_RemoveBlower(sponge)
	end
	
	obstructor = sponge.ssvars.blower.obstructor
	
	// The blower will explode after 30 seconds in non-cooperative gametypes.
	if not G_CoopGametype() then
	
		local timer = blower.cusval
	
		// If something is obstructing the reef blower, add to the explosion timer.
		if (obstructor and obstructor.valid and obstructor.rbflags == RBF_OBSTRUCT) then
			blower.cusval = $ + 1
		else
			blower.cusval = 0
		end
		
		if (timer >= 30*TICRATE) then P_ExplodeBlower(sponge) return end // KABOOM!
	
		// Uh oh, looks like the blower is about to blow!
		if (timer >= 15*TICRATE) then
		
			local x,y
			local smoke
					
			x = P_ReturnThrustX(nil, angle, -sponge.mo.radius*2)
			y = P_ReturnThrustY(nil, angle, -sponge.mo.radius*2)
			
			// In the last twenty seconds, notify the player that it is about to explode.
			if (timer == 20*TICRATE) then
				S_StartSound(blower, sfx_s3kd4s)
			end
			
			// Every 8 seconds, spawn smoke particles behind the player.
			if (timer % 8 == 0) then
				smoke = P_SpawnMobjFromMobj(sponge.mo, x, y, sponge.mo.height/2, MT_SMOKE)
				smoke.scale = P_RandomRange(1/2, 2)*FRACUNIT
				smoke.momz = 5*FRACUNIT
			end
		end
	end
	
	// Something is blocking the Reef Blower!
	if (obstructor and obstructor.valid and obstructor.rbflags != RBF_OBSTRUCT)	
		sponge.ssvars.blower.obstructor.rbflags = RBF_OBSTRUCT // Set the "obstruct" flag, if our obstructor does not currently have one.
	end
	
	// Do not move the nozzle, if our blower no longer exists.
	if not (blower and blower.valid) then
		return
	end
	
	// Set the blower's position in front of the pack's "nozzle"
	P_MoveOrigin(blower, blower.target.x, blower.target.y, blower.target.z)
	P_InstaThrust(blower, angle, 32*FRACUNIT)
	P_Thrust(blower, angle-ANGLE_90, 16*FRACUNIT)
end
addHook("MobjThinker", BlowerThinker, MT_SUPERSPONGE_VACBOX)

addHook("MapChange", function(gamemap)
	gBlowerList = {}
end)

addHook("NetVars", function(net)
	gBlowerList = net($)
end)

// PF_SPINNING removes itself every other frame, so this will have to do.
addHook("PlayerCanDamage", function(player)

	if not (player.mo and player.mo.valid and player.mo.health > 0) then
		return // dead or non-existent
	end
	
	if not (player.mo.rbflags == RBF_LAUNCHED) then
		return // not being launched
	end
	
	if not (player.mo.state == S_PLAY_ROLL) then
		return // not rolling
	end
	
	return true
end)

addHook("MobjMoveBlocked", function(mo, thing, line)

	local blocked
	
	if not (mo.rbflags == RBF_LAUNCHED or mo.rblaunchtime > 0) then
		return
	end
	
	if (line and line.valid and mo.rblaunchtime <= 6) then
		blocked = true
	else
		blocked = false
	end
	
	mo.blocked_by_wall = blocked 
end)

// A small fix which allows objects such as the Pyre Fly to be pulled by the blower.
local function MobjPullFix(mo)
	if (mo.sucker and mo.sucker.valid) then 
		mo.target = nil
		mo.pulled = true
	elseif (mo.pulled) then
		mo.flags2 = $ & ~(MF2_BOSSNOTRAP)
		mo.pulled = false
	end
end
addHook("MobjThinker", MobjPullFix, MT_PYREFLY)