// CLAMS

rawset(_G, "CLAM_LIMIT", 3) // Clam Limit
rawset(_G, "CLAM_ATTACK_DIST", 256*FRACUNIT) // Clam Attack Distance

// 
// P_GetClamCount
//
// Returns the amount of clams in player's array.
//
local function P_GetClamCount(player)
	local count = 0
	
	// No clams are currently active.
	if (player.ssvars.clams == {}) then
		return 0
	end
	
	for i = 1, #(player.ssvars.clams) do
		local clam = player.ssvars.clams[i]
		if not (clam and clam.valid) then 
			continue 
		end
		count = $ + 1
	end
	
	return count
end

// 
// P_DoClamAttack
//
// Clam attacking function
//
rawset(_G, "P_DoClamAttack", function(clam, target)

	// set attack timer
	if not (clam.cusval) then
		clam.cusval = 3*TICRATE
	end

	P_DamageMobj(target, clam, clam.summoner.mo, 1)
end)

//
// P_ClamSearch
//
// It's the function P_LookForEnemies, but for clams! Get it? Heh heh.
//
rawset(_G, "P_ClamSearch", function(clam)

	local mo
	local maxdist = FixedMul(CLAM_ATTACK_DIST, clam.scale)
	local span = ANG30
	local dist, closestdist
	
	dist = 0
	closestdist = 0

	for mobj in mobjs.iterate() do
	
		mo = mobj
	
		if (mo.flags & MF_NOCLIPTHING)
			continue
		end
		
		if (mo.health <= 0) // dead
			continue
		end
		
		if not (mo.flags & (MF_ENEMY|MF_BOSS) and (mo.flags & MF_SHOOTABLE))
			continue // not a valid target
		end
		
		if (mo == clam)
			continue
		end
		
		if (mo.flags2 & MF2_FRET) then
			continue
		end
		
		local zdist = (clam.z + clam.height/2) - (mo.z + mo.height/2)
		dist = R_PointToDist2(0, 0, clam.x - mo.x, clam.y - mo.y)
		
		if ((R_PointToAngle2(0, 0, dist, zdist) + span) > span*2) then
			continue // Don't home outside of desired angle!
		end
		
		dist = R_PointToDist2(0, 0, dist, zdist)
		if (dist > maxdist)
			continue // out of range
		end
		
		if ((twodlevel or clam.flags2 & MF2_TWOD) and abs(clam.y-mo.y) > clam.radius) then
			continue // not in your 2d plane
		end
		
		if (mo.type == MT_PLAYER) 
			continue // TODO: make special case in match?
		end
		
		if ((R_PointToAngle2(clam.x + P_ReturnThrustX(clam, clam.angle, clam.radius), clam.y + P_ReturnThrustY(clam, clam.angle, clam.radius), mo.x, mo.y) - clam.angle + span) > span*2) then
			continue // behind back
		end
		
		if not P_CheckSight(clam, mo) then
			continue // out of sight
		end
		
		return mo
	end
end)

//
// P_UpdateClamArray
//
// Updates the player's clam array
//
rawset(_G, "P_UpdateClamArray", function(player)

	local count

	if (player.ssvars == nil) then
		return // non-existent
	end
	
	count = #(player.ssvars.clams)
	
	if not (count >= 1) then
		return // empty
	end

	// update the array
	for i = 1, #(player.ssvars.clams), 1 do
		local clam = player.ssvars.clams[i]
		if not (clam and clam.valid) then 
			table.remove(player.ssvars.clams, i) 
		else
			clam.id = i
		end
	end
end)

//
// P_SummonClams
//
// "Summons" a clam to the player's aid
//
rawset(_G, "P_SummonClams", function(sponge)

	local clam
	
	if (P_GetClamCount(sponge) >= CLAM_LIMIT) then
		return false
	end
	
	clam = P_SpawnMobjFromMobj(sponge.mo, 0, 0, 0, MT_SUPERSPONGE_CLAM)
	
	if not (clam and clam.valid) then
		return false
	end
	
	// set clam attributes
	clam.angle = sponge.mo.angle
	clam.movedir = (P_RandomChance(FRACUNIT/2) and -1 or 1)
	clam.fuse = 30*TICRATE // clams despawn after 30 seconds
	clam.threshold = 0
	
	// who hast summoned me!?
	if (sponge.mo and sponge.mo.valid) then
		clam.target = sponge.mo
	else
		P_LookForPlayers(clam, true, false, 0)
	end
	
	clam.summoner = clam.target.player // found our summoner
	
	table.insert(clam.summoner.ssvars.clams, clam) // add a clam to the array
	
	clam.id = #(clam.summoner.ssvars.clams) // set id
	
	S_StartSound(clam, sfx_prloop)
	
	return true
end)

//
// P_ClamChase
//
// Modified A_HomingChase for Clams.
//
local function P_ClamChase(clam, speed, target)

	local id
	local x, y, z
	local dest
	local dist
	local angle
	local speedmul

	id = clam.id
	dest = target

	if (not dest or not dest.health)
		return
	end
	
	x = dest.x
	y = dest.y
	z = dest.z
	
	local foffsetx = 0 // X Forward Offset
	local foffsety = 0 // Y Forward Offset
	local boffsetx = 0 // X Sideways Offset
	local boffsety = 0 // Y Sideways Offset
	
	angle = dest.player and dest.player.drawangle or dest.angle
	
	// For the second and third clam, offset horizontal X / Y axis by 16 fracunits.
	if (id == 3) then
		foffsetx = P_ReturnThrustX(clam, angle, -32*FRACUNIT)
		foffsety = P_ReturnThrustY(clam, angle, -32*FRACUNIT)
		boffsetx = P_ReturnThrustX(clam, angle-ANGLE_90, 32*FRACUNIT) // RIGHT
		boffsety = P_ReturnThrustY(clam, angle-ANGLE_90, 32*FRACUNIT)
	elseif (id == 2) then
		foffsetx = P_ReturnThrustX(clam, angle, -32*FRACUNIT)
		foffsety = P_ReturnThrustY(clam, angle, -32*FRACUNIT)
		boffsetx = P_ReturnThrustX(clam, angle-ANGLE_90, -32*FRACUNIT) // LEFT
		boffsety = P_ReturnThrustY(clam, angle-ANGLE_90, -32*FRACUNIT)
	else
		foffsetx = P_ReturnThrustX(clam, angle, -32*FRACUNIT)
		foffsety = P_ReturnThrustY(clam, angle, -32*FRACUNIT)
	end
	
	x = $ + boffsetx
	y = $ + boffsety
	
	// Our source is flipped, so flip destination height as well.
	if (P_MobjFlip(dest) == -1) then
		if not (clam.eflags & MFE_VERTICALFLIP) then
			clam.eflags = $|MFE_VERTICALFLIP
		end
		z = $ - FixedMul(dest.height/32, dest.scale)
	else
		if (clam.eflags & MFE_VERTICALFLIP) then
			clam.eflags = $ & ~MFE_VERTICALFLIP
		end
		z = $ + FixedMul(dest.height/2, dest.scale)
	end

	clam.angle = R_PointToAngle2(clam.x, clam.y, x, y)

	dist = P_AproxDistance(P_AproxDistance(x - clam.x, y - clam.y), z - clam.z)

	if (dist < 1)
		dist = 1
	end

	speedmul = FixedMul(speed, clam.scale)

	clam.momx = FixedMul(FixedDiv(x - clam.x, dist), speedmul)
	clam.momy = FixedMul(FixedDiv(y - clam.y, dist), speedmul)
	clam.momz = FixedMul(FixedDiv(z - clam.z, dist), speedmul)
end

// 
// ClamThink
//
// Clam thinking function
//
local function ClamThink(clam) 

	local id
	local dist
	local angle
	local speed
	local summoner
	
	id = clam.id
	summoner = clam.summoner

	if not (id and clam and clam.valid and summoner and summoner.mo and summoner.mo.valid) then return end
	
	if (clam.state == S_SUPERSPONGE_CLAM_SPIN)
	or (clam.state == S_SUPERSPONGE_CLAM_ATTACK) then
		return // currently attacking or flying off
	end
	
	if (clam.cusval)
		clam.cusval = $ - 1 // Decrement attack timer.
	end
	
	if not (clam.searching) then return end	// We're currently behind our summoner, so ignore this.
	
	dist = P_AproxDistance(P_AproxDistance(summoner.mo.x - clam.x, summoner.mo.y - clam.y), summoner.mo.z - clam.z)
	angle = R_PointToAngle2(clam.x, clam.y, clam.summoner.mo.x, clam.summoner.mo.y)
	
	if (dist <= CLAM_ATTACK_DIST/2) then clam.searching = false return end // we've reached our summoner already
	
	local attack_dist = FixedInt(CLAM_ATTACK_DIST)
	
	// If we've lost sight of our summoner, try teleporting directly to them.
	if not P_CheckSight(clam, summoner.mo) then
	
		if P_MobjFlip(summoner.mo) == -1 then
			P_MoveOrigin(clam, summoner.mo.x, summoner.mo.y, summoner.mo.z - 16*FRACUNIT)		
		else
			P_MoveOrigin(clam, summoner.mo.x, summoner.mo.y, summoner.mo.z + 16*FRACUNIT)
		end
		
		return
	end
	
	// If the summoner speed is greater than our own, try to catch up.
	if (summoner.speed) then
		speed = FixedInt(summoner.speed) + (7 * FixedInt(dist)/attack_dist)
	else
		speed = 32 * FixedInt(dist)/attack_dist
	end
	
	P_ClamChase(clam, speed*FRACUNIT, summoner.mo)
end
addHook("MobjThinker", ClamThink, MT_SUPERSPONGE_CLAM)